<?php
/* Copyright (C) 2021      Open-DSI             <support@open-dsi.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \file        htdocs/custom/banking4dolibarr/class/banking4dolibarrprocess.class.php
 * \ingroup     banking4dolibarr
 * \brief       This file form manage process
 */

/**
 * Class for Banking4DolibarrProcess
 */
class Banking4DolibarrProcess
{
	/**
	 * Is the process supported for the current OS (start ans status)
	 * @return  bool
	 */
	public static function isProcessSupported()
	{
		return preg_match('/^(win|linux)/i', PHP_OS);
	}

	/**
	 * Get process status
	 *
	 * @param	int		$process_id		Process ID
	 * @return  bool
	 */
	public static function getProcessStatus($process_id)
	{
		global $conf;

		if (!empty($process_id) && is_numeric($process_id)) {
			if (preg_match('/^win/i', PHP_OS)) {
				$task_list = array();
				exec("tasklist /FI \"PID eq $process_id\" 2>NUL", $task_list);
				$status = count(array_filter(array_map('trim', $task_list))) > 1;
			} elseif (preg_match('/^linux/i', PHP_OS)) {
				if (!empty($conf->global->BANKING4DOLIBARR_PROCESS_GET_STATUS_METHOD_3) && function_exists('posix_kill')) {
					$status = posix_kill($process_id, 0);
				} elseif (!empty($conf->global->BANKING4DOLIBARR_PROCESS_GET_STATUS_METHOD_2)) {
					$status = shell_exec('ps ' . $process_id . ' | wc -l') > 1;
				} else {
					$status = file_exists('/proc/' . $process_id);
				}
			} else {
				$status = !empty($process_id);
			}
		} else {
			$status = false;
		}

		return $status;
	}

	/**
	 * Start the cli file to process the jobs for a site
	 *
	 * @param   array	$errors			Output errors
	 * @param   string	$script			Command to start in cli
	 * @param   string	$error_log		Filepath of the errors log
	 * @param   array	$params			Parameters to include
	 * @return  int						< 0 if KO otherwise the process ID
	 */
	public static function startProcess(&$errors, $script, $error_log, $params = array())
	{
		dol_syslog(__METHOD__ . " script=$script, error_log=$error_log, params=" . json_encode($params), LOG_DEBUG);
		global $conf, $langs;

		// Clean parameters
		$errors = is_array($errors) ? $errors : array();
		$script = trim($script);
		$error_log = trim($error_log);
		$params = is_array($params) ? $params : array();

		// Check parameters
		if (empty($script)) {
			$langs->load('errors');
			$errors[] = $langs->trans('ErrorBadParameters') . ' - startProcess()';
			dol_syslog(__METHOD__ . " Errors: " . $langs->trans('ErrorBadParameters') . ' - startProcess()', LOG_ERR);
			return -1;
		}

		$langs->load('banking4dolibarr@banking4dolibarr');

		$cmd = trim($conf->global->BANKING4DOLIBARR_PROCESS_START_CMD);
		if (empty($cmd)) {
			if (preg_match('/^win/i', PHP_OS)) {
				$cmd = 'start /B php __SCRIPT__' . (!empty($error_log) ? ' 1>> __ERROR_LOG_FILEPATH__' : '');
			} else {
				$cmd = 'php __SCRIPT__' . (!empty($error_log) ? ' >> __ERROR_LOG_FILEPATH__' : '') . ' & echo $!;';
			}
		}

		if (substr_count($cmd, '__SCRIPT__') != 1) {
			$errors[] = $langs->trans('Banking4DolibarrErrorStartProcessCmd', $cmd);
			dol_syslog(__METHOD__ . " Errors launch process command: " . $langs->trans('Banking4DolibarrErrorStartProcessCmd', $cmd), LOG_ERR);
			return -1;
		}

		// Start process
		$cmd = str_replace('__SCRIPT__', ' ' . escapeshellarg($script) . (!empty($params) ? ' ' . implode(' ', array_map('escapeshellarg', $params)) : '') . ' ', $cmd);
		if (!empty($error_log)) $cmd = str_replace('__ERROR_LOG_FILEPATH__', escapeshellarg($error_log), $cmd);
		$process_errors = array();
		$process_id = self::executeCmd($process_errors, $cmd, !empty($conf->global->BANKING4DOLIBARR_PROCESS_USE_EXEC_METHOD) ? 'exec' : 'proc_open');
		if ($process_id <= 0) {
			$errors[] = $langs->trans('Banking4DolibarrErrorStartProcess');
			if (!empty($process_errors)) $errors = array_merge($errors, $process_errors);
			elseif (!empty($error_log)) $errors[] = 'Read errors file for more details : ' . $error_log;
			dol_syslog(__METHOD__ . " Errors start process (" . $cmd . ") - Result: $process_id - Errors: " . json_encode($process_errors), LOG_ERR);
			return -1;
		}

		return $process_id;
	}

	/**
	 * Execute command in cli
	 *
	 * @param	array		$errors				Output errors
	 * @param	string		$command			Command to execute
	 * @param	string		$method				Execute method ('exec': exec function, 'proc_open' or other: proc_open function)
	 * @param	int			$max_error_length	Max process error string size
	 * @return	int								<=0 if KO, >0 if OK
	 */
	public static function executeCmd(&$errors, $command, $method = 'exec', $max_error_length = 65000)
	{
		dol_syslog(__METHOD__ . " command=$command, method=$method", LOG_DEBUG);
		$process_id = 0;
		$error = 0;

		if ($method == 'exec') {
			$process_id = exec($command, $output, $result_code);
			if ($result_code != 0) {
				$errors[] = implode('; ', $output);
				$error++;
			}
			sleep(1); // For be sure the process is detected as started after this function so don't delete it
		} else {
			$descriptorspec = array(
				0 => array("pipe", "r"),    // stdin est un pipe où le processus va lire
				1 => array("pipe", "w"),    // stdout est un pipe où le processus va écrire
				2 => array("pipe", "w"),    // stderr est un pipe où le processus va écrire
			);
			$cwd = null;
			$env = null;

			$process = proc_open($command, $descriptorspec, $pipes, $cwd, $env);

			// Check if opening has succeed
			if ($process === false || !is_resource($process)) {
				$errors[] = 'Cannot execute process';
				$error++;
			} else {
				sleep(1); // For be sure the process is detected as started so don't delete it

				// Get process PID
				$status = proc_get_status($process);
				if($status === false) {
					$errors[] = 'Failed to obtain status information of the process';
					$error++;
				} else {
					$process_id = $status['pid'] + 1; // To get the real PID of the cmd and not the shell process
					dol_syslog(__METHOD__ . " - PID=" . $status['pid'] . ", True PID=" . $process_id, LOG_DEBUG);
				}

				$process_status = self::getProcessStatus($process_id);
				dol_syslog(__METHOD__ . " - Process status=" . ($process_status ? 1 : 0), LOG_DEBUG);
				if ($process_status) {
					dol_syslog(__METHOD__ . " - Set all streams to non blocking mode", LOG_DEBUG);
					// Set all streams to non blocking mode
					stream_set_blocking ($pipes[0], 0);
					stream_set_blocking ($pipes[1], 0);
					stream_set_blocking ($pipes[2], 0);
				} else {
					dol_syslog(__METHOD__ . " - Get errors", LOG_DEBUG);
					// Get errors
					$process_errors = stream_get_contents($pipes[2]);
					if (!empty($process_errors) || $status['exitcode'] != 0) {
						require_once DOL_DOCUMENT_ROOT . '/core/lib/functions.lib.php';
						if (!empty($process_errors)) $errors[] = dol_trunc($process_errors, $max_error_length); // Max 65000 characters
						$error++;
					}
				}

				// Close pipes
				fclose($pipes[0]);
				fclose($pipes[1]);
				fclose($pipes[2]);

				// Close process
				proc_close($process);
			}
		}

		if ($error) {
			dol_syslog(__METHOD__ . " Errors start process (" . $command . ") - Errors: " . json_encode($errors), LOG_ERR);
			return -1;
		} else {
			return $process_id;
		}
	}
}
