<?php
/* Copyright (C) 2021      Open-DSI             <support@open-dsi.fr>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

/**
 * \file        htdocs/custom/banking4dolibarr/class/banking4dolibarrprocessbox.class.php
 * \ingroup     banking4dolibarr
 * \brief       This file is a class file for manage process box
 */

require_once DOL_DOCUMENT_ROOT . '/core/lib/admin.lib.php';
dol_include_once('/banking4dolibarr/class/banking4dolibarrprocess.class.php', 'Banking4DolibarrProcess');

/**
 * Class for Banking4DolibarrProcessBox
 */
class Banking4DolibarrProcessBox
{
	/**
	 * @var	DoliDB		Database handler.
	 */
	public $db;
	/**
	 * @var	string		Error
	 */
	public $error = '';
	/**
	 * @var	array		Errors
	 */
	public $errors = array();

	/**
	 * @var array   Process info for each object id cached
	 */
	public static $process_info_cached;

	/**
	 * Constructor
	 *
	 * @param DoliDb $db Database handler
	 */
	public function __construct(DoliDB $db)
	{
		$this->db = $db;
	}

	/**
	 * Start the cli file to process the jobs for a object id and a topic
	 *
	 * @param   int		$id				Object ID
	 * @param   string	$topic			Topic
	 * @param   array	$parameters		Parameters
	 * @return  int						< 0 if KO, >0 if OK
	 */
	public function startProcess($id, $topic, $parameters = array())
	{
		global $conf, $langs;
		$result = 1;
		$error = 0;
		$this->errors = array();

		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = trim($topic);
		$parameters = is_array($parameters) ? $parameters : array();

		// Check parameters
		if (empty($id)) {
			$langs->load("errors");
			$this->errors[] = $langs->trans('ErrorFieldRequired', $langs->transnoentitiesnoconv("TechnicalID"));
			$error++;
		}
		if (empty($topic)) {
			$langs->load("errors");
			$this->errors[] = $langs->trans('ErrorFieldRequired', $langs->transnoentitiesnoconv("Banking4DolibarrTopic"));
			$error++;
		}
		if ($error) {
			dol_syslog(__METHOD__ . " Errors check parameters: " . $this->errorsToString(), LOG_ERR);
			return -3;
		}

		$process_list = $this->getProcessInfo($id, $topic);
		if (empty($process_list['status'])) {
			$this->clearEventMessages($id, $topic);

			// Start process
			$error_log = $conf->global->BANKING4DOLIBARR_PROCESS_LOG_FILEPATH;
			if (empty($error_log)) $error_log = DOL_DATA_ROOT . '/banking4dolibarr_process___TOPIC_____ID__.log';
			$error_log = str_replace('__ID__', $id, $error_log);
			$error_log = str_replace('__TOPIC__', $topic, $error_log);
			$script = dol_buildpath('/banking4dolibarr/scripts/process_box.php', 0);
			$process_id = Banking4DolibarrProcess::startProcess($this->errors, $script, $error_log, array($id, $topic, $conf->entity, base64_encode(json_encode($parameters))));
			if ($process_id < 0) {
				dol_syslog(__METHOD__ . " Errors launch process: " . $this->errorsToString(), LOG_ERR);
				return -1;
			}
		}

		return $result;
	}

	/**
	 * Get the process status for a object id and a topic
	 *
	 * @param   int		$id				Object ID
	 * @param   string	$topic			Topic
	 * @return  int						<0 if KO otherwise the status
	 */
	public function statusProcess($id, $topic)
	{
		global $langs;
		$error = 0;
		$this->errors = array();

		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = trim($topic);

		// Check parameters
		if (empty($id)) {
			$langs->load("errors");
			$this->errors[] = $langs->trans('ErrorFieldRequired', $langs->transnoentitiesnoconv("TechnicalID"));
			$error++;
		}
		if (empty($topic)) {
			$langs->load("errors");
			$this->errors[] = $langs->trans('ErrorFieldRequired', $langs->transnoentitiesnoconv("Banking4DolibarrTopic"));
			$error++;
		}
		if ($error) {
			dol_syslog(__METHOD__ . " Errors check parameters: " . $this->errorsToString(), LOG_ERR);
			return -3;
		}

		$process_list = $this->getProcessInfo($id, $topic);

		return $process_list['status'];
	}

	/**
	 * Stop the process for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @return  int					< 0 if KO, >0 if OK
	 */
	public function stopProcess($id, $topic)
	{
		return $this->setStoppingState($id, $topic, 1);
	}

	/**
	 * Load in cache all the process info for all object ids and topics
	 *
	 * @param   bool	$reload		Reload global before get info
	 */
	public function loadProcessInfo($reload = false)
	{
		global $conf;

		if (empty(self::$process_info_cached) || $reload) {
			$process_info_list = array();

			if ($reload) {
				$v17p = version_compare(DOL_VERSION, "17.0.0") >= 0;
				$value = dolibarr_get_const($this->db, 'BANKING4DOLIBARR_PROCESS_INFO', 0);
				if ($v17p) $value = dolDecrypt($value);	// decrypt data excrypted with dolibarr_set_const($db, $name, $value)
				$conf->global->BANKING4DOLIBARR_PROCESS_INFO = $value;
			}

			if (!empty($conf->global->BANKING4DOLIBARR_PROCESS_INFO)) {
				$process_list = explode(',', $conf->global->BANKING4DOLIBARR_PROCESS_INFO);
				foreach ($process_list as $process) {
					$process_info = explode(':', $process);
					$id = trim($process_info[0]);
					$topic = trim($process_info[1]);
					$process_id = trim($process_info[2]);
					$started_at = trim($process_info[3]);
					$stopping = trim($process_info[4]);
					$status = Banking4DolibarrProcess::getProcessStatus($process_id);

					$process_info_list[$id . '_' . $topic] = array (
						'process_id' => $process_id,
						'started_at' => $started_at,
						'status' => !empty($status) ? 1 : 0,
						'stopping' => !empty($stopping) ? 1 : 0,
					);
				}
			}

			self::$process_info_cached = $process_info_list;
		}
	}

	/**
	 * Get the process info for a object id and a topic (process ID, date started, status and stopping)
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @param   bool	$reload		Reload global before get info
	 * @return  array				< 0 if KO otherwise array('process_id' => xxx, 'started_at' => xxx, 'status' => 0 or 1, 'stopping' => 0 or 1)
	 */
	public function getProcessInfo($id, $topic, $reload = false)
	{
		$this->loadProcessInfo($reload);

		if (isset(self::$process_info_cached[$id . '_' . $topic])) {
			return self::$process_info_cached[$id . '_' . $topic];
		}

		return array(
			'process_id' => 0,
			'started_at' => 0,
			'status' => 0,
			'stopping' => 0,
		);
	}

	/**
	 * Set the process info for a object id and a topic (if $process_id == 0 the delete the entry, used for stop the process)
	 *
	 * @param   int		$id					Object ID
	 * @param   string	$topic				Topic
	 * @param   int		$process_id			Process ID
	 * @param   int		$started_at			Process started at
	 * @param   int		$stopping			Stop the process (0: No, 1: Yes)
	 * @param   bool	$reload				Reload global before get info
	 * @return  int							< 0 if KO, >0 if OK
	 */
	public function setProcessInfo($id, $topic, $process_id, $started_at = null, $stopping = 0, $reload = false)
	{
		global $langs;
		$error = 0;
		$this->errors = array();

		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = trim($topic);
		$process_id = $process_id > 0 ? $process_id : 0;
		$started_at = is_numeric($started_at) ? $started_at : null;
		$stopping = $stopping > 0 ? $stopping : 0;

		// Check parameters
		if (empty($id)) {
			$langs->load("errors");
			$this->errors[] = $langs->trans('ErrorFieldRequired', $langs->transnoentitiesnoconv("TechnicalID"));
			$error++;
		}
		if (empty($topic)) {
			$langs->load("errors");
			$this->errors[] = $langs->trans('ErrorFieldRequired', $langs->transnoentitiesnoconv("Banking4DolibarrTopic"));
			$error++;
		}
		if ($error) {
			dol_syslog(__METHOD__ . " Errors check parameters: " . $this->errorsToString(), LOG_ERR);
			return -3;
		}

		$this->loadProcessInfo($reload);
		$process_list = self::$process_info_cached;

		if ($process_id > 0) {
			// Save the process info
			$process_list[$id . '_' . $topic] = array($process_id, $started_at, $stopping);
		} elseif (isset($process_list[$id . '_' . $topic])) {
			// Delete the process info (used for stop set process)
			unset($process_list[$id . '_' . $topic]);
		}
		foreach ($process_list as $id_topic => $process) {
			if (isset($process['status'])) {
				unset($process['status']);
			}
			$process_list[$id_topic] = implode(':', array_merge(array($id, $topic), $process));
		}

		$result = dolibarr_set_const($this->db, 'BANKING4DOLIBARR_PROCESS_INFO', implode(',', $process_list), 'chaine', 0, '', 0);
		if ($result < 0) {
			$this->errors[] = $langs->trans('Banking4DolibarrErrorSaveProcessInfo');
			dol_syslog(__METHOD__ . " Errors save process info: " . $this->errorsToString(), LOG_ERR);
			return -1;
		}

		return 1;
	}

	/**
	 * Set Stopping status for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @param   int		$status		Working status
	 * @return  int					< 0 if KO, >0 if OK
	 */
	public function setStoppingState($id, $topic, $status)
	{
		$process_list = $this->getProcessInfo($id, $topic, true);
		return $this->setProcessInfo($id, $topic, $process_list['process_id'], $process_list['started_at'], $status);
	}

	/**
	 * Get Stopping status for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @return  bool
	 */
	public function getStoppingState($id, $topic)
	{
		$process_list = $this->getProcessInfo($id, $topic, true);
		return !empty($process_list['stopping']) || $process_list['process_id'] != getmypid();
	}

	/**
	 * Clear messages for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 */
	public function clearEventMessages($id, $topic)
	{
		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = strtoupper(trim($topic));

		$const_messages = 'BANKING4DOLIBARR_PROCESS_MESSAGES_' . $id . '_' . $topic;

		dolibarr_del_const($this->db, $const_messages, 0);
	}

	/**
	 * Add message for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @param   string	$message	Message
	 * @param   bool	$print		Print message
	 * @param   string	$type		Type message ('msg', 'error', 'warning')
	 * @param   int		$idx		Number of the current result
	 * @param   int		$max		Max result count
	 * @param   string	$indent		Indent string
	 * @return	string				Message
	 */
	public function addEventMessage($id, $topic, $message, $print = false, $type = 'msg', $idx = 0, $max = 0, $indent = '')
	{
		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = strtoupper(trim($topic));
		$message = trim($message);
		$idx = $idx > 0 ? $idx : 0;
		$max = $max > 0 ? $max : 0;

		if (!empty($message)) {
			if (!empty($max)) {
				$size = strlen($max);
				$message = sprintf("%{$size}d/%d - ", $idx, $max) . $message;
			}
			if (!empty($indent)) $message = $indent . $message;
			if ($type == 'error') $message = '<span style="color: red;">' . $message . '</span>';
			elseif ($type == 'warning') $message = '<span style="color: orange;">' . $message . '</span>';

			$const_messages = 'BANKING4DOLIBARR_PROCESS_MESSAGES_' . $id . '_' . $topic;
			$messages = $this->getEventMessages($id, $topic);
			$messages[] = $message;

			// Remove firsts messages if json to long for TEXT format size (65535)
			$messages = $this->checkSizeMessages($messages);

			dolibarr_set_const($this->db, $const_messages, json_encode($messages), 'chaine', 0, '', 0);

			if ($print) {
				return dol_escape_htmltag(dol_htmlentitiesbr_decode($message)) . "\n";
			}
		}

		return $message;
	}

	/**
	 * Check messages size
	 *
	 * @param   array		$messages			Messages
	 * @return  array
	 */
	public function checkSizeMessages($messages)
	{
		// Remove first message if json to long for TEXT format size (65535)
		$value = json_encode($messages);
		$size = dol_strlen($value);
		if ($size > 65535) {
			$messages = array_slice($messages, 1);
			return $this->checkSizeMessages($messages);
		}

		return $messages;
	}

	/**
	 * Get messages for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @return  array
	 */
	public function getEventMessages($id, $topic)
	{
		global $conf;

		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = strtoupper(trim($topic));

		$const_messages = 'BANKING4DOLIBARR_PROCESS_MESSAGES_' . $id . '_' . $topic;
		$messages = empty($conf->global->$const_messages) ? array() : json_decode($conf->global->$const_messages, true);

		return $messages;
	}

	/**
	 * Set post data for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @param   array	$data		Post data
	 */
	public function setProcessPostData($id, $topic, $data)
	{
		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = strtoupper(trim($topic));

		$const_messages = 'BANKING4DOLIBARR_PROCESS_POST_DATA_' . $id . '_' . $topic;

		dolibarr_set_const($this->db, $const_messages, json_encode($data), 'chaine', 0, '', 0);
	}

	/**
	 * Get post data for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 * @param   bool	$clear		Clear also post data
	 * @return  array
	 */
	public function getProcessPostData($id, $topic, $clear = false)
	{
		global $conf;

		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = strtoupper(trim($topic));

		$const_messages = 'BANKING4DOLIBARR_PROCESS_POST_DATA_' . $id . '_' . $topic;
		$post_data = empty($conf->global->$const_messages) ? array() : json_decode($conf->global->$const_messages, true);

		if ($clear) $this->clearProcessPostData($id, $topic);

		return $post_data;
	}

	/**
	 *  Load setup values into conf object (read llx_const) for a specified entity
	 *  Note that this->db->xxx, this->file->xxx and this->multicompany have been already loaded when setValues is called.
	 *
	 * @param	int		$entity		Entity to get
	 * @return	int					< 0 if KO, >= 0 if OK
	 */
	function setEntityValues($entity)
	{
		global $conf;

		$result = 0;

		if ($conf->entity != $entity) {
			$conf->entity = $entity;

			if (method_exists($conf, 'setEntityValues')) {
				$result = $conf->setEntityValues($this->db, $entity);
			} else {
				// Unset all old modules values
				if (!empty($conf->modules)) {
					foreach ($conf->modules as $m) {
						if (isset($conf->$m)) unset($conf->$m);
					}
				}

				// Properly declare multi-modules objects.
				$conf->global = new stdClass();
				$conf->multicompany = new stdClass();

				// First level object
				// TODO Remove this part.
				$conf->expedition_bon = new stdClass();
				$conf->livraison_bon = new stdClass();
				$conf->fournisseur = new stdClass();
				$conf->product = new stdClass();
				$conf->service = new stdClass();
				$conf->contrat = new stdClass();
				$conf->actions = new stdClass();
				$conf->commande = new stdClass();
				$conf->propal = new stdClass();
				$conf->facture = new stdClass();
				$conf->contrat = new stdClass();
				$conf->usergroup = new stdClass();
				$conf->adherent = new stdClass();
				$conf->bank = new stdClass();
				$conf->notification = new stdClass();
				$conf->mailing = new stdClass();
				$conf->expensereport = new stdClass();
				$conf->productbatch = new stdClass();
				$conf->modules = array();;
				$conf->modules_parts = array(
					'css' => array(),
					'js' => array(),
					'tabs' => array(),
					'triggers' => array(),
					'login' => array(),
					'substitutions' => array(),
					'menus' => array(),
					'theme' => array(),
					'sms' => array(),
					'tpl' => array(),
					'barcode' => array(),
					'models' => array(),
					'societe' => array(),
					'hooks' => array(),
					'dir' => array(),
					'syslog' => array(),
				);

				$result = $conf->setValues($this->db);
			}
		}

		return $result;
	}

	/**
	 * Clear post data for a object id and a topic
	 *
	 * @param   int		$id			Object ID
	 * @param   string	$topic		Topic
	 */
	public function clearProcessPostData($id, $topic)
	{
		// Clean parameters
		$id = $id > 0 ? $id : 0;
		$topic = strtoupper(trim($topic));

		$const_messages = 'BANKING4DOLIBARR_PROCESS_POST_DATA_' . $id . '_' . $topic;

		dolibarr_del_const($this->db, $const_messages, 0);
	}

	/**
	 * Method to output saved errors
	 *
	 * @param   string      $separator      Separator between each error
	 * @return	string		                String with errors
	 */
	public function errorsToString($separator = ', ')
	{
		return $this->error . (is_array($this->errors) ? (empty($this->error) ? '' : $separator) . join($separator, $this->errors) : '');
	}
}
